from datetime import datetime
from hypothesis import assume
from sentenai import V, hql
from sentenai.api.stream import Stream
from sentenai.historiQL import merge

def stream(name, *filters):
    return Stream(None, name, {}, None, True, *filters)

def test_merge_construction():
    s = stream("1")
    evt = s.foo == 1

    span1 = hql.And(evt, hql.Within(days=1))
    span2 = hql.And(evt, hql.Within(days=2))
    assume( merge(span1, span2) )




from sentenai.stream.metadata import Metadata
from sentenai.api import *
if PANDAS:
    import pandas as pd
from datetime import datetime, time, date
import simplejson as JSON
import re, io, math
from collections import namedtuple
from multiprocessing import Pool
from tqdm import tqdm, tqdm_notebook
from time import sleep
import cbor2
from shapely.geometry import Point

from queue import Queue
from threading import Thread
from concurrent.futures import ThreadPoolExecutor



def worker(q, workers, total, progress, position=None):
    with ThreadPoolExecutor(max_workers=workers) as pool:
        if progress:
            v = 0
            with tqdm(total=total, unit=" values", position=position) as pbar:
                while True:
                    data = q.get()
                    if not data:
                        pbar.update(total - v)
                        break # exit on empty list
                    list(pool.map(index_data, data))
                    n = sum([len(v) for d, n, i, v in data])
                    v += n
                    pbar.update(n)
        else:
            while True:
                data = q.get()
                if not data:
                    break # exit on empty list
                list(pool.map(index_data, data))


def index_data(args):
    db, node, index, v = args
    counter = 0
    while counter < 10:
        try:
            data = cbor2.dumps(v)
        except Exception as e:
            print(e)
            print(db, node, index)
            print(v[-1])

        try:
            resp = db._post('nodes', node, 'types', index,
                    json=data, headers={'Content-Type': 'application/cbor'}, raw=True)
        except:
            counter += 1
            sleep(.1)
        else:
            if resp.status_code > 204:
                resp.close()
                raise Exception(f"failed on index")
            resp.close()
            break


class Database(API):
    def __init__(self, parent, name, origin):
        self._parent = parent
        self._name = name
        self._origin = origin
        API.__init__(self, parent._credentials, *parent._prefix, "db", name)



    def __iter__(self):
        r = self._get("links")
        if r.status_code != 200:
            raise SentenaiError("invalid response")
        data = r.json()
        return iter(sorted(data.keys()))

    def keys(self):
        return iter(self)
   
    def items(self):
        return iter([(k, self[k]) for k in self])

    def values(self):
        return iter([self[k] for k in self])

    def graph(self, path=None, limit=-1):
        import treelib
        t = treelib.Tree()
        root = t.create_node(path[-1] if path else self.name, path[-1] if path else self.name, data={})
        ps = {}
        if limit >= 0:
            ps['limit'] = limit
        r = self._get("graph", *(path or []), params=ps)
        if r.status_code != 200:
            raise SentenaiError("Invalid Response")
        data = r.json()
        for node in sorted(data, key=lambda x: x[0]):
            parent = root
            x = [path[-1] if path else self.name]
            for link in node[0][len(path) - 2 if path else 0:]:
                x.append(link)
                nid = "/".join(x)
                if nid in t:
                    continue
                else:
                    pid = "/".join(x[:-1]) 
                    t.create_node(link, nid, parent=pid, data={'id': node[1], 'type': node[2], 'children': node[3], 'indexes': node[4]})
        return t


    @property
    def meta(self):
        raise NotImplemented("metadata not implemented on databases.")
        return Metadata(self)

    #def load(self, file):
    #    self._post(file)

    def __getitem__(self, key):
        if isinstance(key, tuple):
            return Stream(self, *key)
        else:
            return Stream(self, key)

    def __setitem__(self, key, content):
        workers = 32
        chunksize = 4096
        if isinstance(key, tuple):
            if isinstance(key[-1], slice):
                path = key[:-1]
                path += key[-1].start
                workers = key[-1].stop
                chunksize = key.step or chunksize
            else:
                path = key
        elif isinstance(key, slice):
            workers = key.stop
            chunksize = key.step or chunksize
            path = (key.start,)
        else:
            path = (key,)

        del self[path]

        if content is None:
            nid = self._put('paths', *path, json={'kind': 'directory'})
            return
        elif PANDAS and isinstance(content, pd.DataFrame):
            nid = self._put('paths', *path).json()['node']
            self._put('nodes', nid, 'types', 'event')
            cmap = {'start': nid}
            tmap = {'start': 'event'}
            dmap = {'start': []}
            df = content.sort_values(by='start', ignore_index=True)
        elif type(content) == str:
            nid = self._put('paths', *path, json={'kind': 'virtual', 'tspl': content})
            return
        elif content == float:
            nid = self._put('paths', *path).json()['node']
            self._put('nodes', nid, 'types', 'float')
            return
        elif content == int:
            nid = self._put('paths', *path).json()['node']
            self._put('nodes', nid, 'types', 'int')
            return
        elif content == bool:
            nid = self._put('paths', *path).json()['node']
            self._put('nodes', nid, 'types', 'bool')
            return
        elif content == datetime:
            nid = self._put('paths', *path).json()['node']
            self._put('nodes', nid, 'types', 'datetime')
            return
        elif content == date:
            nid = self._put('paths', *path).json()['node']
            self._put('nodes', nid, 'types', 'date')
            return
        elif content == time:
            nid = self._put('paths', *path).json()['node']
            self._put('nodes', nid, 'types', 'time')
            return
        elif content == str:
            nid = self._put('paths', *path).json()['node']
            self._put('nodes', nid, 'types', 'text')
            return
        elif isinstance(content, Stream):
            nid = content._node
            if isinstance(key, tuple) and len(key) > 1:
                src = self[key[:-1]]
                self._put('nodes', src._node, 'links', key[-1], nid)
            else:
                if isinstance(key, tuple) and len(key) > 1:
                    key = key[0]
                self._put('links', key, nid)
            return
        elif isinstance(content, list):
                cmap = {}
                tmap = {}
                dmap = {}
                df = pd.DataFrame(content).rename(columns={'value': path[-1]})
                if set(df.columns) == {'start', 'end', path[-1]}:
                    pass
                elif set(df.columns) == {'start', 'end'}:
                    dmap = {'start': []}
                else:
                    raise Exception(str(df.columns))
                df = df.sort_values(by='start', ignore_index=True)
        else:
            raise TypeError("invalid assignment type")

        if len(df) == 0:
            raise ValueError("Cannot index empty dataset")

        def add(cname):
            retries = 100
            while retries > 0:
                try:
                    if isinstance(content, list):
                        nid = self._put('paths', *path).json()['node']
                    else:
                        nid = self._put('paths', *path, cname).json()['node']
                    if df[cname].dtype == np.dtype('float32'):
                        tm = 'float'
                    elif df[cname].dtype == np.dtype('float64'):
                        tm = 'float'
                    elif df[cname].dtype == np.dtype('int32'):
                        tm = 'int'
                    elif df[cname].dtype == np.dtype('int64'):
                        tm = 'int'
                    elif df[cname].dtype == bool:
                        tm = 'bool'
                    elif df[cname].dtype == np.dtype('datetime64[ns]'):
                        tm = 'datetime'
                    elif df[cname].dtype == np.dtype('timedelta64[ns]'):
                        tm = 'timedelta'
                    elif type(df[cname][0]) == date:
                        tm = 'date'
                    elif type(df[cname][0]) == time:
                        tm = 'time'
                    elif type(df[cname][0]) == Point and df[cname][0].has_z:
                        tm = 'point3'
                    elif type(df[cname][0]) == Point:
                        tm = 'point'
                    else:
                        tm = 'text'

                    self._put('nodes', nid, 'types', tm)
                    #cmap[cname] = nid
                    #dmap[cname] = []
                    #tmap[cname] = tm
                    return (cname, nid, tm)
                except Exception as e:
                    retries -= 1
                    sleep(0.5)
            raise Exception("failed to create node/index")


        with ThreadPoolExecutor(max_workers=1) as pool:
            res = pool.map(add, [x for x in df.columns if x not in ['start', 'end']])
            for cname, nid, tm in res:
                cmap[cname] = nid
                dmap[cname] = []
                tmap[cname] = tm


        origin = self.origin
        res = []
        q = Queue()
        wt = Thread(target=worker, args=(q, workers, len(df) * (len(df.columns) - 1), self._parent.interactive))
        wt.start()

        for i, row in df.iterrows():
            if origin is not None:
                ts = int((dt64(row['start']) - origin) // np.timedelta64(1, 'ns'))
            else:
                try:
                    ts = int(td64(row['start']) // np.timedelta64(1, 'ns'))
                except:
                    print(row['start'], row['start'].dtype)
                    raise
            try:
                if 'duration' in row:
                    dur = int(row['duration'])
                elif 'end' in row and origin is not None:
                    dur = int((dt64(row['end']) - dt64(row['start'])) // np.timedelta64(1, 'ns'))
                elif origin is not None:
                    dur = int((dt64(df['start'].iloc[i+1]) - dt64(row['start'])) // np.timedelta64(1, 'ns'))
                elif 'end' in row:
                    dur = int((td64(row['end']) - td64(row['start'])) // np.timedelta64(1, 'ns'))
                else:
                    dur = int((td64(df['start'].iloc[i+1]) - td64(row['start'])) // np.timedelta64(1, 'ns'))
            except IndexError:
                dur = 1

            if dur <= 0: continue
            for col, val in dict(row).items():
                if col == 'start' and 'start' in dmap:
                    dmap[col].append((ts, dur))
                elif type(val) == float and math.isnan(val): # skip nans
                    pass
                elif val is pd.NaT or val is None:
                    pass
                elif col not in tmap:
                    pass
                elif tmap[col] == 'point3':
                    dmap[col].append((ts, dur, (val.x, val.y, val.z)))
                elif tmap[col] == 'point':
                    dmap[col].append((ts, dur, (val.x, val.y)))
                elif tmap[col] == 'date':
                    dmap[col].append((ts, dur, val.isoformat()))
                elif tmap[col] == 'time':
                    dmap[col].append((ts, dur, val.isoformat()))
                elif tmap[col] == 'datetime':
                    dmap[col].append((ts, dur, iso8601(val)))
                elif tmap[col] == 'timedelta':
                    dmap[col].append((ts, dur, val // np.timedelta64(1, 'ns')))
                else:
                    dmap[col].append((ts, dur, val))

            if 'start' in dmap and len(dmap['start']) >= chunksize:
                q.put([(self, cmap[k], tmap[k], dmap[k]) for k, v in dmap.items()])
                for k, v in dmap.items():
                    dmap[k] = []
            elif len(list(dmap.values())[0]) >= chunksize:
                q.put([(self, cmap[k], tmap[k], dmap[k]) for k, v in dmap.items()])
                for k, v in dmap.items():
                    dmap[k] = []

        if 'start' in dmap and len(dmap['start']) > 0:
            q.put([(self, cmap[k], tmap[k], dmap[k]) for k, v in dmap.items()])
            for k, v in dmap.items():
                dmap[k] = []
        elif len(list(dmap.values())[0]) > 0:
            q.put([(self, cmap[k], tmap[k], dmap[k]) for k, v in dmap.items()])
            for k, v in dmap.items():
                dmap[k] = []
        q.put([])
        wt.join()
        
            


    def __delitem__(self, key):
        if isinstance(key, tuple):
            self._delete('paths', *key)
        else:
            self._delete('paths', key)

    def __repr__(self):
        return f"Database({self._parent!r}, \"{self._name}\")"

    def __str__(self):
        return str(self._name)

    @property
    def origin(self):
        return self._origin

    @property
    def name(self):
        return self._name

class Node(API):
    def __init__(self, db, node):
        self._node = node
        API.__init__(self, db._credentials, *db._prefix, "nodes", self._node)

    @property
    def meta(self):
        return Metadata(self)

class Stream(API):
    def __init__(self, parent, *path):
        self._parent = parent
        self._path = path
        r = self._parent._get('paths', *path)
        if r.status_code == 404:
            raise KeyError("path does not exist")
        else:
            self._node = r.json()['node']
        API.__init__(self, parent._credentials, *parent._prefix, "nodes", self._node)

    def __setitem__(self, key, v):
        if not isinstance(key, tuple):
            key = (key,)

        self._parent[self._path + key] = v

    @property
    def links(self):
        return iter(self)

    @property
    def children(self):
        return iter((self[k] for k in self.links))


    def __iter__(self):
        r = self._get("links")
        if r.status_code != 200:
            raise SentenaiError("invalid response")
        data = r.json()
        return iter(sorted(data.keys()))

    def insert(self, values):
        vs = []
        origin = self._parent.origin
        if origin is not None:
            for v in values:
                start = (dt64(v['start']) - origin) // np.timedelta64(1, 'ns')
                end = (dt64(v['end']) - origin) // np.timedelta64(1, 'ns')
                if 'value' in v:
                    vs.append((int(start), int(end - start), v['value']))
                else:
                    vs.append((int(start), int(end - start)))
        else:
            for v in values:
                start = td64(v['start']) // np.timedelta64(1, 'ns')
                end = td64(v['end']) // np.timedelta64(1, 'ns')
                if 'value' in v:
                    vs.append((int(start), int(end - start), v['value']))
                else:
                    vs.append((int(start), int(end - start)))
        self._post('types', self.type,
                json=cbor2.dumps(vs), headers={'Content-Type': 'application/cbor'}, raw=True)

    def export(self, start=None, end=None, limit=None, exclude=tuple(), origin=datetime(1970,1,1), when=None):
        exp = API(self._credentials, "export")
        o = iso8601(self._parent.origin or origin)[:-1] + 'Z'
        co = ["start", "end"]
        cols = [f"{self}/{x}" for x in iter(self) if x not in exclude]
        when = when or str(self)
        params = {}
        if limit is not None:
            params['limit'] = limit
        if start is not None:
            params['start'] = iso8601(start) if o else int(start)
        if end is not None:
            params['end'] = iso8601(end) if o else int(end)
        if o is not None:
            params['origin'] = o

        r = exp._post(json={'when': when, 'select': co+cols}, params=params)
        df = pd.DataFrame( r.json(), columns = co + [x for x in list(self) if x not in exclude])
        df.start = df.start.apply(pd.Timestamp)
        df.end = df.end.apply(pd.Timestamp)
        return df

    def graph(self, limit=-1):
        return self._parent.graph(self._path, limit)

    @property
    def source(self):
        r = self._get()
        if r.status_code == 200:
            info = r.json()
            if not info:
                return None
            return info.get('tspl')
        else:
            return None



    @property
    def meta(self):
        return Metadata(self)

    @meta.setter
    def meta(self, md):
        payload = {}
        for key, val in md.items():
            if isinstance(val, bool):
                vtype = "bool"
            elif isinstance(val, datetime) or isinstance(val, np.datetime64):
                vtype = "datetime"
                val = dt64(val)
            elif isinstance(val, int):
                vtype = "int"
            elif isinstance(val, float):
                vtype = "float"
            else:
                vtype = "text"
            payload[key] = {'type': vtype, 'value': val}
        resp = self._patch('meta', json=payload)
        if resp.status_code not in [200, 201, 204]:
            raise Exception(resp.status_code)


    @property
    def type(self):
        r = self._get('types')
        if r.status_code == 200:
            ts = r.json()
            print(ts)
            if not ts:
                return None
            else:
                return ts[0]
        else:
            return None

    @property
    def range(self):
        if self.type is None:
            print("NO INDEXES")
            return None
        r = self._get('types', self.type, 'range')
        if r.status_code == 200:
            e = r.json()
            if e is None:
                return None
            if self._parent.origin is None:
                return (e['start'], e['end'])
            else:
                return (self._parent.origin + np.timedelta64(e['start'], 'ns'), self._parent.origin + np.timedelta64(e['end'], 'ns'))
        else:
            return None

    def __repr__(self):
        z = ", ".join(map(repr, self._path))
        return f"Stream({self._parent!r}, {z})"

    def __str__(self):
        return "/".join((self._parent.name,) + self._path)

    def __getitem__(self, key):
        if isinstance(key, tuple):
            return Stream(self._parent, *(self._path + key))
        else:
            return Stream(self._parent, *(self._path + (key,)))
  
    @property
    def raw(self):
        return RawData(self)


    @property
    def data(self):
        return StreamData(self, self.type, False)

    @property
    def df(self):
        return StreamData(self, self.type, True)

    @property
    def first(self):
        try:
            return self.data[::1][0]
        except KeyError:
            return None

    def head(self, n=20, df=True):
        if n <= 0: return None
        if df:
            return self.df[::n]
        else:
            return self.data[::n]
    
    def tail(self, n=20, df=True):
        if n <= 0: return None
        if df:
            return self.df[::-n]
        else:
            return self.data[::-n]

    @property
    def last(self):
        try:
            return self.data[::-1][0]
        except KeyError:
            return None
    @property
    def stats(self):
        return StreamStats(self)

    #def __len__(self):
    #    return StreamStats(self, vtype=self.type, ro=True).count
    
class StreamStats(object):
    def __init__(self, parent, start=None, end=None, origin=None, vtype=None, ro=False):
        self._parent = parent
        self._start = start
        self._end = end
        self._type = vtype
        self._origin = origin
        self._read_only = ro

    def __getitem__(self, tr):
        if self._read_only:
            raise TypeError("Stream stats already has a time range")

        if isinstance(tr, tuple):
            if len(tr) == 2:
                s, o = tr
                t = self._type
            elif len(tr) == 3:
                s, o, t = tr
            else:
                raise ValueError("invalid arguments to .data")
            o = iso8601(o)
        else:
            s = tr
            o = self._origin
            t = self._type

        if s.start is not None:
            self._start = s.start
        if s.stop is not None:
            self._end = s.stop

        self._read_only = True
        return self


    def _stat(self, stat):
        origin = self._origin or self._parent._parent.origin or dt64("1970-01-01T00:00:00")
        start, end = self._parent.range

        #start = self._start or origin
        #end = self._end or origin + td64(2 ** 63 - 1)
        try:
            return self._parent._parent._parent(f"""
                {stat}({"when " if self._type == 'event' else ''}{self._parent})
                    when interval({iso8601(start)}, {iso8601(end)})
                    origin {iso8601(origin)}
            """)[::1][0]['value']
        except BadRequest:
            raise TypeError from None
        except IndexError:
            return None
        except KeyError:
            return None

    @property
    def count(self):
        return self._stat('count')

    @property
    def mean(self):
        return self._stat('mean')

    @property
    def min(self):
        return self._stat('min')

    @property
    def max(self):
        return self._stat('max')

    @property
    def sum(self):
        return self._stat('sum')

    @property
    def std(self):
        return self._stat('std')

    @property
    def top(self):
        return self._stat('top')

    @property
    def any(self):
        return self._stat('any')

    @property
    def all(self):
        return self._stat('all')

    
class StreamData(API):
    def __init__(self, parent, index, df=False, resample=None, rolling=None, origin=''):
        self._parent = parent
        self._df = df
        self._type = index
        self._origin = origin
        self._resample = resample
        self._rolling = rolling
        API.__init__(self, parent._credentials, *parent._prefix, "types", index)

    def origin(self, o=None):
        if not o:
            return StreamData(self._parent, self._type, self._df, self._resample, self._rolling, '')
        else:
            return StreamData(self._parent, self._type, self._df, self._resample, self._rolling, 'origin ' + iso8601(o))

    def resample(self, period, aggregator=None):
        if self._type == 'event':
            aggregator = "count"
        return StreamData(self._parent, self._type, self._df, (period, aggregator), self._rolling)

    def rolling(self, period):
        return StreamData(self._parent, self._type, self._df, self._resample, (period, "trailing"))


    def __getitem__(self, tr):
        if self._rolling:
            r = f'window {self._rolling[0]} {self._rolling[1]}'
        else:
            r = ''
        if self._resample:
            if self._resample[1]:
                rs = f'{self._resample[1]}({self._parent!s}) when frequency({self._resample[0]}) {r}'
            else:
                rs = f'{self._parent!s} when frequency({self._resample[0]}) {r}'
        else:
            rs = f'{"when " if self._type == "event" else ""}{self._parent!s} {r}'

        if self._df:
            return self._parent._parent._parent.df(rs + self._origin)[tr]
        else:
            return self._parent._parent._parent(rs + self._origin)[tr]



class RawData(API):
    def __init__(self, parent):
        self._parent = parent

    def __getitem__(self, i):
        params = {}
        if isinstance(i, slice):

            if i.start is None:
                pass
            elif type(i.start) is int:
                params['start'] = int(i.start)
            else:
                params['start'] = iso8601(i.start)

            if i.stop is None:
                pass
            elif type(i.stop) is int:
                params['end'] = int(i.stop)
            else:
                params['end'] = iso8601(i.stop)

            if i.step is not None:
                params['limit'] = i.step
        if self._parent.type == 'event':
            resp = self._parent._parent._parent._post('tspl', json=f'when {self._parent!s}', params=params, headers={'Accept': 'application/cbor'})
        else:
            resp = self._parent._parent._parent._post('tspl', json=str(self._parent), params=params, headers={'Accept': 'application/cbor'})
        if 'content-type' in resp.headers and resp.headers['content-type'] == 'application/cbor':
            return cbor2.loads(resp.content)
        else:
            return resp.json()






from datetime import datetime
from enum import Enum, auto
from sentenai.api import API, Credentials, iso8601
import time
from dataclasses import dataclass
from typing import Optional
import numpy as np
import cbor2
import uuid, io
import pathlib


class APIException(Exception): pass
class IndexNotFound(APIException): pass



@dataclass
class Event:
    offset: np.timedelta64
    duration: np.timedelta64

    def start(self, origin: Optional[np.datetime64] = None):
        if origin:
            return origin + self.offset
        else:
            return self.offset

    def end(self, origin: Optional[np.datetime64] = None):
        if origin:
            return origin + self.offset + self.duration
        else:
            return self.offset + self.duration

@dataclass
class Float(Event):
    value: float

@dataclass
class Int(Event):
    value: int

@dataclass
class Bool(Event):
    value: bool

@dataclass
class Text(Event):
    value: str

@dataclass
class Point(Event):
    value: (float, float)

@dataclass
class Point3(Event):
    value: (float, float, float)


class Tempest(object):
    def __init__(self, host='localhost', port=7280):
        self.host = host
        self.port = port
        self.protocol = 'http://'

    @property
    def api(self):
        h = f"{self.protocol}{self.host}:{self.port}"
        return API(Credentials(h, None))

    def __getitem__(self, k):
        r = tempest.api._get('db', k)
        return DB(k)


tempest = Tempest()


def databases():
    return [DB(db) for db in sorted(tempest.api._get('db').json())]

class DB:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name

    @property
    def links(self):
        return Links(self)

    @property
    def paths(self):
        return Paths(self)


    def init(self, origin=datetime(1970,1,1)):
        if origin == None:
            r = tempest.api._put("db", self.name, json={'origin': None})
        else:
            r = tempest.api._put("db", self.name, json={'origin': iso8601(origin)})
        if r.status_code != 201:
            raise Exception("Could not initialize")


epoch = datetime(1970, 1, 1)

CACHE_DURATION = 1

class TSPL:
    def __init__(self, src):
        self._src = src
        self._explain = None


    def __getitem__(self, s):
        params = {}
        if s.start:
            params['start'] = s.start
        if s.stop:
            params['end'] = s.stop
        if s.step:
            params['limit'] = s.step
        return tempest.api._post('tspl', json=self._src, params=params).json()

    @property
    def explain(self):
        if not self._explain:
            self._explain = tempest.api._post('tspl/debug', json=self._src).json()
        return self._explain

    @property
    def range(self):
        return tempest.api._post('tspl/range', json=self._src).json()

    @property
    def tsvm(self):
        return ""

    @property
    def mermaid(self):
        return ""

    def diagram(self):
        from IPython.display import IFrame
         
        def js_ui(data, template, out_fn = None, out_path='.',
                  width="800px", height="600px", **kwargs):
            """Generate an IFrame containing a templated javascript package."""
            if not out_fn:
                out_fn = pathlib.Path(f"{uuid.uuid4()}.html")
                 
            # Generate the path to the output file
            out_path = pathlib.Path(out_path)
            filepath = out_path / out_fn
            # Check the required directory path exists
            filepath.parent.mkdir(parents=True, exist_ok=True)
         
            # The open "wt" parameters are: write, text mode;
            with io.open(filepath, 'wt', encoding='utf8') as outfile:
                # The data is passed in as a dictionary so we can pass different
                # arguments to the template
                outfile.write(template.format(**data))
         
            return IFrame(src=filepath, width=width, height=height)
        
        return js_ui({'src': self.explain['mermaid']}, TEMPLATE_MERMAIDJS)


def tspl(src):
    return TSPL(src)



class Kind(Enum):
    Directory = 'directory'
    Indexed = 'indexed'
    View = 'view'
    Virtual = 'view'

class Path:
    kind = Kind.Indexed

    @classmethod
    def json(cls):
        return {'kind': cls.kind.value}

class Indexed(Path):
    kind = Kind.Indexed

class Directory(Path):
    kind = Kind.Directory

class Virtual(Path):
    def __init__(self, tspl):
        self.source = tspl
    kind = Kind.Virtual


class Type(Enum):
    Int = 'int'
    Float = 'float'
    Bool = 'bool'
    Point = 'point'
    Point3 = 'point3'
    Text = 'text'
    Event = 'event'
    Date = 'date'
    Time = 'time'
    DateTime = 'datetime'
    TimeDelta = 'timedelta'

class Paths:
    def __init__(self, db):
        self.db = db

    def __getitem__(self, *path):
        resp = tempest.api._get('db', self.db.name, 'paths', *path).json()
        return Node(self.db, resp['node'])
    
    def __setitem__(self, path, p):
        tempest.api._put("db", self.db.name, 'paths', path, json=p.json())
        


class Node(object):
    def __init__(self, db, nid):
        self.db = db
        self.id = nid

    def __repr__(self):
        return f'Node("{self.id}")'

    @property
    def links(self):
        return Links(self)

    @property
    def types(self):
        return Types(self)

    @property
    def meta(self):
        return Meta(self)

class Meta(dict):
    def __init__(self, node):
        self.node = node
        self._meta = {}
        self._age = 0

    @property
    def _md(self):
        now = time.time()
        if now - self._age > CACHE_DURATION:
            self._meta = {k: v['value'] for k, v in tempest.api._get('db', self.node.db.name, 'nodes', self.node.id, 'meta').json().items()}
        return self._meta

    def __setitem__(self, key, item):
        raise NotImplemented

    def __getitem__(self, key):
        return self._md[key]

    def __repr__(self):
        return repr(self._md)

    def __len__(self):
        return repr(self._md)

    def __delitem__(self):
        raise NotImplemented
    
    def clear(self):
        raise NotImplemented

    def copy(self):
        return Links(self.node, self._md)

    def has_key(self, key):
        return key in self._md
    
    def update(self, *args, **kwargs):
        raise NotImplemented

    def keys(self):
        return self._md.keys()
    
    def values(self):
        return self._md.values()
    
    def values(self):
        return self._md.items()
    
    def pop(self, *args):
        return self._md.pop(*args)

    def __cmp__(self, dict_):
        return cmp(self._md, dict_)

    def __contains__(self, item):
        return item in self._md
    
    def __iter__(self):
        return iter(self._md)

    def __unicode__(self):
        return unicode(repr(self._md))




class Stream(object):
    def __init__(self, node, type):
        self.type = type
        self.node = node

    def __len__(self):
        resp = tempest.api._head('db', self.node.db.name, 'nodes', self.node.id, 'types', self.type.value)
        if resp.status_code == 200:
            return int(resp.headers['Count'])
        else:
            raise IndexNotFound


    def insert(self, *data):
        cbor = None
        cbor = cbor2.dumps([(int(np.timedelta64(v.offset, 'ns')), int(np.timedelta64(v.duration, 'ns')), v.value) for v in data])
        """
        if self.type == Type.Float:
            cbor = cbor2.dumps([(int(np.timedelta64(v.offset, 'ns')), int(np.timedelta64(v.duration, 'ns')), float(v.value)) for v in data])
        else:
            if isinstance(v.value, Text):
                cbor = cbor2.dumps([(int(np.timedelta64(v.offset, 'ns')), int(np.timedelta64(v.duration, 'ns')), v.value) for v in data])
            elif isinstance(v.value, Bool):
                cbor = cbor2.dumps([(int(np.timedelta64(v.offset, 'ns')), int(np.timedelta64(v.duration, 'ns')), v.value) for v in data])


            cbor = cbor2.dumps([(int(np.timedelta64(v.offset, 'ns')), int(np.timedelta64(v.duration, 'ns')), tuple(map(float, v.value))) for v in data])
        """
        resp = tempest.api._post('db', self.node.db.name, 'nodes', self.node.id, 'types', self.type.value,
                    json=cbor, headers={'Content-Type': 'application/cbor'}, raw=True)

    def __getitem__(self, params):
        resp = tempest.api._get('db', self.node.db.name, 'nodes', self.node.id, 'types', self.type.value,
                    headers={'Accept': 'application/cbor'}, params=params)
        if self.type == Type.Event:
            return [Event(np.timedelta64(x['ts'], 'ns'), np.timedelta64(x['duration'], 'ns')) for x in resp.json()]
        elif self.type == Type.Float:
            return [Float(np.timedelta64(x['ts'], 'ns'), np.timedelta64(x['duration'], 'ns'), x['value']) for x in resp.json()]
        elif self.type == Type.Text:
            return [Text(np.timedelta64(x['ts'], 'ns'), np.timedelta64(x['duration'], 'ns'), x['value']) for x in resp.json()]
        elif self.type == Type.Bool:
            return [Bool(np.timedelta64(x['ts'], 'ns'), np.timedelta64(x['duration'], 'ns'), x['value']) for x in resp.json()]
        elif self.type == Type.Point:
            return [Point(np.timedelta64(x['ts'], 'ns'), np.timedelta64(x['duration'], 'ns'), tuple(x['value'])) for x in resp.json()]
        elif self.type == Type.Point3:
            return [Point3(np.timedelta64(x['ts'], 'ns'), np.timedelta64(x['duration'], 'ns'), tuple(x['value'])) for x in resp.json()]

    @property 
    def range(self):
        resp = tempest.api._get('db', self.node.db.name, 'nodes', self.node.id, 'types', self.type.value, 'range')
        return resp.json()
        

class Types(dict):
    def __init__(self, node):
        self.node = node
        self._types = {}
        self._age = 0

    def __call__(self, type):
        return Stream(self.node, type)

    def attach(self, type):
        resp = tempest.api._put('db', self.node.db.name, 'nodes', self.node.id, 'types', type.value)
        return Stream(self.node, type)


    @property
    def _ts(self):
        now = time.time()
        if now - self._age > CACHE_DURATION:
            ls = tempest.api._get('db', self.node.db.name, 'nodes', self.node.id, 'types').json()
            self._types = {Type(k) : Stream(self.node, Type(k)) for k in ls}
        return self._types

    def __setitem__(self, key, item):
        raise NotImplemented

    def __getitem__(self, key):
        return self._ts[key]

    def __repr__(self):
        return repr(self._ts)

    def __len__(self):
        return repr(self._ts)

    def __delitem__(self):
        raise NotImplemented
    
    def clear(self):
        raise NotImplemented

    def copy(self):
        return Links(self.node, self._ts)

    def has_key(self, key):
        return key in self._ts
    
    def update(self, *args, **kwargs):
        raise NotImplemented

    def keys(self):
        return self._ts.keys()
    
    def values(self):
        return self._ts.values()
    
    def values(self):
        return self._ts.items()
    
    def pop(self, *args):
        return self._ts.pop(*args)

    def __cmp__(self, dict_):
        return cmp(self._ts, dict_)

    def __contains__(self, item):
        return item in self._ts
    
    def __iter__(self):
        return iter(self._ts)

    def __unicode__(self):
        return unicode(repr(self._ts))

class Links(dict):
    def __init__(self, node):
        self.node = node
        self._links = {}
        self._age = 0

    @property
    def _ls(self):
        now = time.time()
        if False or now - self._age > CACHE_DURATION:
            if isinstance(self.node, DB):
                ls = tempest.api._get('db', self.node.name, 'links').json()
                self._links = {k : Node(self.node, nid) for k, nid in ls.items()}
            else:
                ls = tempest.api._get('db', self.node.db.name, 'nodes', self.node.id, 'links').json()
                self._links = {k : Node(self.node.db, nid) for k, nid in ls.items()}
        return self._links

    def __setitem__(self, key, item):
        raise NotImplemented

    def __getitem__(self, key):
        return self._ls[key]

    def __repr__(self):
        return repr(self._ls)

    def __len__(self):
        return repr(self._ls)

    def __delitem__(self, name):
        if isinstance(self.node, DB):
            tempest.api._delete('db', self.node.name, 'links', name)
        else:
            tempest.api._delete('db', self.node.db, 'nodes', self.node.id, 'links', name)
    
    def clear(self):
        raise NotImplemented

    def copy(self):
        return Links(self.node, self._ls)

    def has_key(self, key):
        return key in self._ls
    
    def update(self, *args, **kwargs):
        raise NotImplemented

    def keys(self):
        return self._ls.keys()
    
    def values(self):
        return self._ls.values()
    
    
    def pop(self, *args):
        return self._ls.pop(*args)

    def __cmp__(self, dict_):
        return cmp(self._ls, dict_)

    def __contains__(self, item):
        return item in self._ls
    
    def __iter__(self):
        return iter(self._ls)

    def items(self):
        return self._ls.items()

    def __unicode__(self):
        return unicode(repr(self._ls))



TEMPLATE_MERMAIDJS="""<html>
    <head>
            <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.3.0/css/all.min.css">
    <head>
    <body>
        <script type="module">
            import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@10/dist/mermaid.esm.min.mjs';
            mermaid.initialize({{ startOnLoad: true }});
        </script>
 
        <pre class="mermaid">
            {src}
        </pre>
 
    </body>
</html>
"""



from sentenai.view import Views, View
import pytest


def test_views(client):
    assert isinstance(client.views, Views)


def test_view_create_and_get(client):
    v = client.views.create('foo', {
        'a': client.streams['weather']['temperatureMax']
    })

    assert client.views['foo'].name == v.name


def test_view_delete(client):
    del client.views['foo']
    with pytest.raises(KeyError):
        client.views['foo']


def test_view_anon(client):
    v = client.views(a=client.streams['weather']['temperatureMax'])
    assert isinstance(v, View)

@pytest.fixture(scope="module")
def vdata(client):
    v = client.views.create('test-weather', {"a": client.streams['weather']['temperatureMax']})
    return v.data

def test_view_data_limit(client, vdata):
    t0 = client.streams['weather'].events[0].ts
    t1 = client.streams['weather'].events[-1].ts
    x = vdata[t0:t1:10]
    assert isinstance(x, list) and len(x) == 10

def test_view_data_limit_0(client, vdata):
    t0 = client.streams['weather'].events[0].ts
    t1 = client.streams['weather'].events[-1].ts
    x = vdata[t0:t1:0]
    assert isinstance(x, list) and len(x) == 0

def test_view_data_limit_neg(client, vdata):
    t0 = client.streams['weather'].events[0].ts
    t1 = client.streams['weather'].events[-1].ts
    with pytest.raises(ValueError):
        x = vdata[t0:t1:-10]




# coding=utf-8
import pytest
from sentenai import Sentenai, hql, V
from sentenai.historiQL import ast_dict, Returning, Or, Lasting, Within, After, Select
from sentenai.hql import stream

def test_basic_select_span():
    s = stream("S")
    real = ast_dict(s.x == True)
    expected = {
        "select": {
            "type": "span",
            "op": "==",
            "stream": { "name": "S" },
            "path": ( "event", "x" ),
            "arg": { "type": "bool", "val": True }
        }
    }

    assert real == expected

# TODO: figure out date/time/datetime types
def test_any_of_comparisons():
    s = stream("moose")
    real = ast_dict(hql.Any(
        s.x < 0,
        s.x >= 3.141592653589793,
        s.b != False
    ))

    expected = {
        "select": {
            "type": "any",
            "conds": [
                {
                    "stream": { "name": "moose" },
                    "arg": { "type": "double", "val": 0 },
                    "path": ( "event", "x" ),
                    "type": "span",
                    "op": "<"
                },
                {
                    "stream": { "name": "moose" },
                    "arg": { "type": "double", "val": 3.141592653589793 },
                    "path": ( "event", "x" ),
                    "type": "span",
                    "op": ">="
                },
                {
                    "stream": { "name": "moose" },
                    "arg": { "type": "bool", "val": False },
                    "path": ( "event", "b" ),
                    "type": "span",
                    "op": "!="
                }
            ]
        }
    }

    assert real == expected

def test_stream_access():
    s = stream("S")
    real = ast_dict(
        s.even == True,
        s.event == True,
        s.event.event == True,
        s.id == True,
        s['.id'] == True,
        s['.id'][''] == True,
        s['true']['真实'] == True
    )
    expected = {
        "select": {
            "type": "serial",
            "conds": [
                {
                    "type": "span",
                    "op": "==",
                    "arg": { "type": "bool", "val": True },
                    "path": ( "event", "even" ),
                    "stream": { "name": "S" }
                },
                {
                    "type": "span",
                    "op": "==",
                    "arg": { "type": "bool", "val": True },
                    "path": ( "event", "event" ),
                    "stream": { "name": "S" }
                },
                {
                    "type": "span",
                    "op": "==",
                    "arg": { "type": "bool", "val": True },
                    "path": ( "event", "event", "event" ),
                    "stream": { "name": "S" }
                },
                {
                    "type": "span",
                    "op": "==",
                    "arg": { "type": "bool", "val": True },
                    "path": ( "event", "id" ),
                    "stream": { "name": "S" }
                },
                {
                    "type": "span",
                    "op": "==",
                    "arg": { "type": "bool", "val": True },
                    "path": ( "event", ".id" ),
                    "stream": { "name": "S" }
                },
                {
                    "type": "span",
                    "op": "==",
                    "arg": { "type": "bool", "val": True },
                    "path": ( "event", ".id", "" ),
                    "stream": { "name": "S" }
                },
                {
                    "type": "span",
                    "op": "==",
                    "arg": { "type": "bool", "val": True },
                    "path": ( "event", "true", "真实" ),
                    "stream": { "name": "S"  }
                }
            ]
        }
    }

    assert real == expected

def test_all_any_serial():
    foo = stream("foo")
    bar = stream("bar")
    baz = stream("baz")
    qux = stream("qux")
    quux = stream("quux")

    real = ast_dict(
        hql.Any(foo.x == True, bar.y == True),
        baz.z == True,
        hql.All(qux['α'] == True, quux['β'] == True)
    )
    expected = {
        "select": {
            "type":"serial",
            "conds": [
                {
                    "type": "any",
                    "conds": [
                        {"op":"==", "stream": {"name": "foo"}, "path":("event","x"), "type":"span", "arg":{"type":"bool", "val":True}},
                        {"op":"==", "stream": {"name": "bar"}, "path":("event","y"), "type":"span", "arg":{"type":"bool", "val":True}}
                    ]
                },
                {
                    "op":"==", "stream": {"name": "baz"}, "path":("event","z"), "type":"span", "arg":{"type":"bool", "val":True}
                },
                {
                    "type": "all",
                    "conds": [
                        {"op":"==", "stream": {"name": "qux"},  "path":("event","α"), "type":"span", "arg":{"type":"bool", "val":True}},
                        {"op" :"==", "stream": {"name": "quux"}, "path":("event","β"), "type":"span", "arg":{"type":"bool", "val":True}}
                    ]
                }
            ]
        }
    }
    assert real == expected

def test_and():
    s = stream('s')
    t = stream('t')
    real = ast_dict(s.x == True & t.x == True)
    expected = {
        "select": {
            "expr": "&&",
            "args": [
               { "type": "span", "op": "==", "stream": {"name": "s"}, "path":("event","x"), "arg": {"type":"bool", "val":True} },
               { "type": "span", "op": "==", "stream": {"name": "t"}, "path":("event","x"), "arg": {"type":"bool", "val":True} }
            ]
        }
    }
    assert real == expected

def test_or():
    s = stream('s')
    t = stream('t')
    real = ast_dict(s.x == True | t.x == True)
    expected = {
        "select": {
            "expr": "||",
            "args": [
               { "type": "span", "op": "==", "stream": {"name": "s"}, "path":("event","x"), "arg": {"type":"bool", "val":True} },
               { "type": "span", "op": "==", "stream": {"name": "t"}, "path":("event","x"), "arg": {"type":"bool", "val":True} }
            ]
        }
    }
    assert real == expected

def test_order_of_operations():
    s = stream('s')
    real = ast_dict(s.x == 1 | s.y > 2 & s.z <= 3)
    expected = {
        "select": {
            "expr": "||",
            "args": [
                {
                    "type": "span",
                    "op": "==",
                    "stream": {"name": "s"},
                    "path":("event","x"),
                    "arg": {"type":"double", "val":1}
                },
                {
                    "expr": "&&",
                    "args": [
                        {
                            "type": "span",
                            "op": ">",
                            "stream": {"name": "s"},
                            "path":("event","y"),
                            "arg": {"type":"double", "val":2}
                        },
                        {
                            "type": "span",
                            "op": "<=",
                            "stream": {"name": "s"},
                            "path":("event","z"),
                            "arg": {"type":"double", "val":3}
                        }
                    ]
                }
            ]
        }
    }
    assert real == expected

def test_parens():
    s = stream('s')
    real = ast_dict((s.x == 1 | s.y > 2) & s.z <= 3)
    expected = {
        'select': {
            'expr': '&&',
            'args': [
                {
                    'expr': '||',
                    'args': [
                        {
                            'path': ('event', 'x'),
                            'arg': {'type': 'double', 'val': 1},
                            'type': 'span',
                            'stream': {'name': 's'},
                            'op': '=='
                        },
                        {
                            'path': ('event', 'y'),
                            'arg': {'type': 'double', 'val': 2},
                            'type': 'span',
                            'stream': {'name': 's'},
                            'op': '>'
                        }
                    ]
                },
                {
                    'path': ('event', 'z'),
                    'arg': {'type': 'double', 'val': 3},
                    'type': 'span',
                    'stream': {'name': 's'},
                    'op': '<='
                }
            ]
        }
    }
    assert real == expected

def test_relative_span():
    s = stream('s')
    t = stream('t')
    real = ast_dict(
        Or(
            (s.x == True, Lasting.min(years=1, months=1) ),
            (t.x == True, Within(seconds=13), After(minutes=11))
        ),
        Lasting.max(days=7)
    )
    expected = {
        "select": {
            "expr": "||",
            "args": [
                {
                    "type": "span",
                    "op": "==",
                    "stream": {"name": "s"},
                    "path": ("event","x"),
                    "arg": {"type":"bool", "val":True},
                    "for": { "at-least": { "years": 1, "months": 1 } }
                },
                {
                    "type": "span",
                    "op": "==",
                    "stream": {"name": "t"},
                    "path": ("event","x"),
                    "arg": {"type":"bool", "val":True},
                    "after": {"minutes": 11},
                    "within": {"seconds": 13}
                }
            ],
            "for": { "at-most": { "days": 7 } }
        }
    }
    assert real == expected

def test_nested_relative_spans():
    s = stream('S')
    real = ast_dict(
        (s.x < 0) >> (
            ((s.x == 0) >> (s.x > 0, Within(seconds=1))),
            Within(seconds=2)
        )
    )
    expected = {
        "select": {
            "type": "serial",
            "conds": [
                {
                    "type":"span",
                    "op":"<",
                    "stream":{"name":"S"},
                    "path":("event","x"),
                    "arg":{"type":"double", "val":0}
                },
                {
                    "type": "serial",
                    "conds": [
                        {
                            "type":"span",
                            "op":"==",
                            "stream":{"name":"S"},
                            "path":("event","x"),
                            "arg":{"type":"double", "val":0}
                        },
                        {
                            "type":"span",
                            "op":">",
                            "stream":{"name":"S"},
                            "path":("event","x"),
                            "arg":{"type":"double", "val":0},
                            "within": {"seconds":1}
                        }
                    ],
                    "within": {"seconds":2}
                }
            ]
        }
    }
    assert real == expected

def test_stream_filters():
    s = stream('S', V.season == "summer")
    real = ast_dict(s.temperature >= 77 & s.sunny == True)
    expected = {
        "select": {
            "expr": "&&",
            "args": [
              {"type": "span", "op": ">=",
               "stream": {"name": "S", "filter": {"op":"==", "path":("event","season"), "arg":{"type":"string","val":"summer"}}},
               "path":("event","temperature"), "arg": {"type":"double", "val":77}
              },
              {"type": "span", "op": "==",
               "stream": {"name": "S", "filter": {"op":"==", "path":("event","season"), "arg":{"type":"string","val":"summer"}}},
               "path":("event","sunny"), "arg": {"type":"bool", "val":True}
              }
            ]
        }
    }
    assert real == expected

def test_or_stream_filters():
    s = stream('S', (V.season == "summer") | (V.season == "winter"))
    real = ast_dict(s.sunny == True)
    expected = {
        'select': {
            'type': 'span',
            'op': '==',
            'arg': {'type': 'bool', 'val': True},
            'path': ('event', 'sunny'),
            'stream': {
                'name': 'S',
                'filter': {
                    'expr': '||',
                    'args': [
                        {'op': '==', 'arg': {'type': 'string', 'val': 'summer'}, 'path': ('event', 'season')},
                        {'op': '==', 'arg': {'type': 'string', 'val': 'winter'}, 'path': ('event', 'season')}
                    ]
                }
            }
        }
    }
    assert real == expected

# def test_switches():
#     s = stream('S')
#     real = ast_dict(
#         select().span(s(event(V.x < 0) >> event(V.x > 0)))
#     )
#     expected = {
#         "select": {
#             "type": "switch",
#             "stream": { "name": "S" },
#             "conds": [
#                 {
#                     "op": "<",
#                     "arg": { "type": "double", "val": 0 },
#                     "path": ( "event", "x" )
#                 },
#                 {
#                     "op": ">",
#                     "arg": { "type": "double", "val": 0 },
#                     "path": ( "event", "x" )
#                 }
#             ]
#         }
#     }
#     assert real == expected

# def test_unary_switch():
#     s = stream('S')
#     real = ast_dict(
#         select().span(s(event(V.x < 0)))
#     )
#     expected = {
#         "select": {
#             "type": "switch",
#             "stream": { "name": "S" },
#             "conds": [
#                 {'expr': True},
#                 {
#                     "op": "<",
#                     "arg": { "type": "double", "val": 0 },
#                     "path": ( "event", "x" )
#                 }
#             ]
#         }
#     }
#     assert real == expected

def test_returning():
    s = stream('weather')
    real = (Returning(s % {
        'value': V.maxTemp,
        'other': {
            'constant': 3
        }
    }))()
    expected = {
        'projections': {
            'explicit': [{
                'stream': {'name': 'weather'},
                'projection': {
                    'value': [{'var': ('event', 'maxTemp')}],
                    'other': {'constant': [{'lit': {'val': 3, 'type': 'int'}}]}
                }
            }],
            '...': True
        }
    }
    assert real == expected

def test_returning_excluding():
    s = stream('weather')
    real = (
        Returning(-s)
    )()
    expected = {
        'projections': {
            'explicit': [{
                'stream': {'name': 'weather'},
                'projection': False
            }],
            '...': True
        }
    }
    assert real == expected

def test_during():
    s = stream('S')
    real = ast_dict(
        hql.During(
            s.foo == 'bar',
            s.baz > 1.5
        )
    )
    expected = {
        "select": {
            "type": "during",
            "conds": [{
                "op": "==",
                "arg": {
                    "type": "string",
                    "val": "bar"
                },
                "type": "span",
                "path": (
                    "event",
                    "foo"
                ),
                "stream": {
                    "name": "S"
                }
            }, {
                "op": ">",
                "arg": {
                    "type": "double",
                    "val": 1.5
                },
                "type": "span",
                "path": (
                    "event",
                    "baz"
                ),
                "stream": {
                    "name": "S"
                }
            }]
        }
    }
    assert real == expected



import pytest
from uuid import uuid4
from sentenai.stream import Stream

def test_streams_iter(client):
    """tests iterability as dict of Streams object."""
    assert dict(client.streams) or True


def test_streams_get(client):
    """Tests default getitem of Streams object."""
    n = uuid4().hex
    s = client.streams[n]
    assert isinstance(s, Stream)
    assert s.name == n
    assert bool(s) == False


def test_delete_nonexistent_stream(client):
    n = uuid4().hex
    with pytest.raises(KeyError):
        del client.streams[n]




import math
from copy import copy
from datetime import datetime, timedelta
import numpy as np
from sentenai.api import API, dt64, td64, iso8601

import collections

def flatten(d, parent_key='', sep='_'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)

class Updates(API):
    def __init__(self, parent):
        API.__init__(self, parent._credentials, *parent._prefix, "events", params=parent._params)
        self._parent = parent

    def __iter__(self):
        i = 0
        n = 100
        while True:
            e = self[i::n]
            if len(e) == 0:
                raise StopIteration
            else:
                for x in e:
                    yield e
                i += n

    def __delitem__(self, i):
        res = self._delete(i)
        if res.status_code == 404:
            raise KeyError("Event does not exist.")
        elif res.status_code != 204:
            raise Exception(res.status_code)

    def __len__(self):
        res = self._head(params=self._params)
        return int(res.headers['events'])

    def __setitem__(self, key, event):
        event.id = key
        self.insert(event)

    def __getitem__(self, i):

        if isinstance(i, str):
            # this is get by id
            res = self._get(i)
            if res.status_code == 200:
                ej = res.json()
                try:
                    ts = int(res.headers['timestamp'])
                except:
                    ts = res.headers['timestamp']
                return Event(id=i, ts=ts, duration=res.headers.get('duration'), data=ej)
            elif res.status_code == 404:
                raise KeyError("Updates does not exist")
            else:
                raise Exception(res.status_code)

        params = copy(self._params)

        if isinstance(i, int):
            raise TypeError("Integer not valid")

        elif isinstance(i, slice):
            # time slice
            t0 = self._parent.t0
            params['sort'] = 'asc'
            if i.start is not None:
                if t0 is None:
                    params['start'] = int(i.start)
                else:
                    params['start'] = iso8601(i.start)
            if i.stop is not None:
                if t0 is None:
                    params['end'] = int(i.stop)
                else:
                    params['end'] = iso8601(i.stop)
            if i.step is not None:
                params['limit'] = abs(i.step)
                if i.step < 0:
                    params['sort'] = 'desc'
            if i.start is not None and i.stop is not None:
                if i.start > i.stop:
                    params['start'], params['end'] = params['end'], params['start']
                    params['sort'] = 'desc'

            resp = self._get(params=params)
            if resp.status_code == 200:
                evts = []
                for ej in resp.json():
                    try:
                        ts = int(ej['ts'])
                    except:
                        ts = ej['ts']
                    evts.append(
                        Event(
                            id=ej['id'],
                            ts=ts,
                            duration=ej.get("duration"),
                            data=ej['event'] or None
                        )
                    )
                return evts
            else:
                raise Exception(resp.status_code)
        else:
            raise ValueError("input must be either string or slice")

    def update(self, evt):
        hdrs = {}
        if evt.id is None:
            raise ValueError("Event id required for updates.")
        if evt.ts is not None:
            hdrs["timestamp"] = iso8601(dt64(evt.ts))
        if evt.duration is not None:
            hdrs["duration"] = str(td64(evt.duration).astype(float) / 1000000000.)
        self._put(evt.id, json=evt.data, headers=hdrs)

    def insert(self, evt):
        hdrs = {'content-type': 'application/json'}
        if evt.ts is not None and evt.duration is None:
            hdrs["timestamp"] = iso8601(evt.ts)
        elif evt.duration is not None:
            hdrs['start'] = iso8601(evt.ts)
            hdrs["end"] = iso8601(evt.ts + evt.duration)

        if evt.id is not None:
            r = self._put(evt.id, json=evt.data, headers=hdrs)
            if r.status_code in [200, 201]:
                return evt
                #return self[r.headers['Location']]
            else:
                raise Exception(r.status_code)
        else:
            r = self._post(json=evt.data, headers=hdrs)
            if r.status_code in [200, 201]:
                return Event(id=r.headers['Location'], data=evt.data, ts=evt.ts, duration=evt.duration)
                #return self[r.headers['Location']]
            else:
                raise Exception(r.status_code)

    def remove(self, evt):
        del self[evt.id]


